# -*- coding: utf-8 -*-
# File: pipe.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains pipeline components that can be plugged into each other and predictors that are invoked by their
 respective component.
"""

from .anngen import *
from .base import *
from .cell import *
from .common import *
from .concurrency import *
from .doctectionpipe import *
from .language import *
from .layout import *
from .lm import *
from .refine import *
from .registry import *
from .segment import *
from .text import *
from .transform import *
